/*
 * Copyright (c) 2008 Jesper Karlsson & David Gustafsson
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <gtk/gtk.h>

#include "xfbib-list-store.h"
#include "xfbib-entry-tree-view.h"
#include "xfbib-bibtex.h"
#include "xfbib-state.h"
#include "xfbib-bibtex-entry.h"
#include "xfbib-entry-edit-dialog.h"

struct _XfbibEntryTreeView
{
	GtkTreeView parent;
};

typedef struct _XfbibEntryTreeViewClass
{
	GtkTreeViewClass parent;
} XfbibEntryTreeViewClass;

static void xfbib_entry_tree_view_class_init(XfbibEntryTreeViewClass *klass);

static void xfbib_entry_tree_view_init(XfbibEntryTreeView *instance);
static void xfbib_entry_tree_view_finalize(GObject *obj);

static GObjectClass *xfbib_entry_tree_view_parent_class = NULL;

static void
cb_row_activated (GtkTreeView *tree_view, GtkTreePath *path, GtkTreeViewColumn *col, gpointer data)
{
	g_print("cb_row_activated\n");
	XfbibState *state;
	XfbibBibtexEntry *entry, *new_entry;
	GtkWidget *dialog;
	GtkTreeModel *model;
	GtkTreeIter iter;

	state = xfbib_state_new();
	model = gtk_tree_view_get_model(tree_view);

	if (gtk_tree_model_get_iter(model, &iter, path)) {
		entry = XFBIB_BIBTEX_ENTRY(xfbib_list_store_get(XFBIB_LIST_STORE(model), &iter));
		dialog = xfbib_entry_edit_dialog_new(entry);
		for(;;) {
			if (gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_OK) {
				if(xfbib_entry_edit_dialog_valid_entry(XFBIB_ENTRY_EDIT_DIALOG(dialog))) {
					new_entry = xfbib_entry_edit_dialog_get_entry(XFBIB_ENTRY_EDIT_DIALOG(dialog));
					xfbib_list_store_replace(XFBIB_LIST_STORE(model), G_OBJECT(entry), G_OBJECT(new_entry));
					break;
				}
			} else {
				break;
			}
		}
		gtk_widget_destroy(dialog);
	}
}

static gboolean
cb_button_press (GtkWidget *tree_view, GdkEventButton *event, gpointer data)
{
	g_print("cb_button_press\n");
	XfbibBibtexEntry *new_entry;
	GtkTreeModel *model;
	GtkWidget *dialog;

	if (event->type == GDK_BUTTON_PRESS && event->button == 3) {
		/* single click with the right mouse button */
		/* TODO: Add a context menu here */
	} else if (event->type == GDK_2BUTTON_PRESS && event->button == 1) {
		/* double click with the left mouse button */

		model = gtk_tree_view_get_model(GTK_TREE_VIEW(tree_view));

		/* Check if any row has been selected */
		if (gtk_tree_selection_get_selected (gtk_tree_view_get_selection
					(GTK_TREE_VIEW(tree_view)), NULL, NULL)) {
			/* 
			 * A row is selected Let the other cb_row_activated()
			 * function take care of the rest.
			 */
			/* TODO: Take care of the event here instead */
			return FALSE;
		}

		dialog = xfbib_entry_edit_dialog_new(NULL);

		for(;;) {
			if (gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_OK) {
				if(xfbib_entry_edit_dialog_valid_entry(XFBIB_ENTRY_EDIT_DIALOG(dialog))) {
					new_entry = xfbib_entry_edit_dialog_get_entry(XFBIB_ENTRY_EDIT_DIALOG(dialog));
					xfbib_list_store_add(XFBIB_LIST_STORE(model), G_OBJECT(new_entry));
					break;
				}
			} else {
				break;
			}
		} 
		gtk_widget_destroy(dialog);

	} else {
		return FALSE;
	}
	return TRUE;
}


GType
xfbib_entry_tree_view_get_type (void)
{
	static GType type = 0;
	if (type == 0) {
		static const GTypeInfo info = {
			sizeof (XfbibEntryTreeViewClass),
			NULL,   /* base_init */
			NULL,   /* base_finalize */
			(GClassInitFunc) xfbib_entry_tree_view_class_init,   /* class_init */
			NULL,   /* class_finalize */
			NULL,   /* class_data */
			sizeof (XfbibEntryTreeView),
			0,      /* n_preallocs */
			(GInstanceInitFunc) xfbib_entry_tree_view_init    /* instance_init */
		};
		type = g_type_register_static (GTK_TYPE_TREE_VIEW, "XfbibEntryTreeView", &info, 0);
	}
	return type;
}


static void
xfbib_entry_tree_view_class_init(XfbibEntryTreeViewClass *klass)
{
	GObjectClass *object_class = (GObjectClass *)klass;

	xfbib_entry_tree_view_parent_class = g_type_class_peek_parent (klass);

	object_class->finalize = xfbib_entry_tree_view_finalize;
}

static void
xfbib_entry_tree_view_init(XfbibEntryTreeView *instance)
{
	GtkTreeViewColumn *column;
	GtkCellRenderer *renderer;
	GtkTreeModel *model;
	gint i;
	struct {
			gchar *label;
			gchar *tooltip;
	}xfbib_field_constants[] = {
			{"Address", _("Address of publisher")},
			{"Annote", _("Annotation for annotated bibliography styles")},
			{"Author", _("Name(s) of the author(s), separated by 'and' if more than one")},
			{"Booktitle", _("Title of the book, if only part of it is being cited")},
			{"Chapter", _("Chapter number")},
			{"Crossref", _("Citation key of the cross-referenced entry")},
			{"Edition", _("Edition of the book (such as \"first\" or \"second\")")},
			{"Editor", _("Name(s) of the editor(s), separated by 'and' if more than one")},
			{"E-print", _("Specification of electronic publication")},
			{"HowPublished", _("Publishing method if the method is nonstandard")},
			{"Institution", _("Institution that was involved in the publishing")},
			{"Journal", _("Journal or magazine in which the work was published")},
			{"Key", _("Hidden field used for specifying or overriding the alphabetical order of entries")},
			{"Month", _("Month of publication or creation if unpublished")},
			{"Note", _("Miscellaneous extra information")},
			{"Number", _("Number of journal, magazine, or tech-report")},
			{"Organization", _("Sponsor of the conference")},
			{"Pages", _("Page numbers separated by commas or double-hyphens")},
			{"Publisher", _("Name of publisher")},
			{"School", _("School where thesis was written")},
			{"Series", _("Series of books in which the book was published")},
			{"Title", _("Title of the work")},
			{"Type", _("Type of technical report")},
			{"URL", _("Internet address")},
			{"Volume", _("Number of the volume")},
			{"Year", _("Year of publication or creation if unpublished")}
	};

	/* Connect the double-click on a row to a callback function */
	g_signal_connect(GTK_TREE_VIEW(instance), "row-activated", (GCallback) cb_row_activated, NULL);
	/* Connect the double-click NOT on a row to a callback function */
	g_signal_connect(GTK_WIDGET(instance), "button-press-event", (GCallback) cb_button_press, NULL);
	
	for (i = 0; i < XFBIB_BIBTEX_FIELD_N_FIELDS; i++) {
		switch (i) {
			case XFBIB_BIBTEX_FIELD_ANNOTE:
			case XFBIB_BIBTEX_FIELD_BOOKTITLE:
			case XFBIB_BIBTEX_FIELD_CHAPTER:
			case XFBIB_BIBTEX_FIELD_CROSSREF:
			case XFBIB_BIBTEX_FIELD_EDITION:
			case XFBIB_BIBTEX_FIELD_EDITOR:
			case XFBIB_BIBTEX_FIELD_EPRINT:
			case XFBIB_BIBTEX_FIELD_HOWPUBLISHED:
			case XFBIB_BIBTEX_FIELD_INSTITUTION:
			case XFBIB_BIBTEX_FIELD_JOURNAL:
			case XFBIB_BIBTEX_FIELD_KEY:
			case XFBIB_BIBTEX_FIELD_MONTH:
			case XFBIB_BIBTEX_FIELD_NOTE:
			case XFBIB_BIBTEX_FIELD_NUMBER:
			case XFBIB_BIBTEX_FIELD_ORGANIZATION:
			case XFBIB_BIBTEX_FIELD_PAGES:
			case XFBIB_BIBTEX_FIELD_PUBLISHER:
			case XFBIB_BIBTEX_FIELD_SCHOOL:
			case XFBIB_BIBTEX_FIELD_SERIES:
			case XFBIB_BIBTEX_FIELD_TYPE:
			case XFBIB_BIBTEX_FIELD_URL:
			case XFBIB_BIBTEX_FIELD_VOLUME:
				continue;
				break;
		}
		column = gtk_tree_view_column_new();

		gtk_tree_view_column_set_title(column, xfbib_field_constants[i].label);

		/* pack tree view column into tree view */
		gtk_tree_view_append_column(GTK_TREE_VIEW(instance), column);

		renderer = gtk_cell_renderer_text_new();

		/* pack cell renderer into tree view column */
		gtk_tree_view_column_pack_start(column, renderer, TRUE);

		/*
		 * connect 'text' property of the cell renderer to
		 * model column that contains the first name
		 */
		gtk_tree_view_column_add_attribute(column, renderer, "text", i);
	}

	model = GTK_TREE_MODEL(xfbib_list_store_new(TYPE_ENTRY));
	gtk_tree_view_set_model(GTK_TREE_VIEW(instance), model);
	g_object_unref(model); /* destroy model automatically with view */
}

static void
xfbib_entry_tree_view_finalize(GObject *obj)
{
	G_OBJECT_CLASS(xfbib_entry_tree_view_parent_class)->finalize(obj);
}

GtkWidget *
xfbib_entry_tree_view_new()
{
	XfbibEntryTreeView *tree_view;
	tree_view = g_object_new(XFBIB_TYPE_ENTRY_TREE_VIEW, NULL);
	return GTK_WIDGET(tree_view);
}

