/*
 * Copyright (c) 2008 Jesper Karlsson & David Gustafsson
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <gtk/gtk.h>
#include <libxfce4ui/libxfce4ui.h>
#include <stdlib.h>

#include "xfbib-file-io.h"
#include "xfbib-state.h"
#include "xfbib-bibtex.h"
#include "xfbib-bibtex-field.h"
#include "xfbib-bibtex-entry.h"
#include "xfbib-bibtex-comment.h"
#include "xfbib-bibtex-preamble.h"
#include "xfbib-bibtex-string.h"
#include "xfbib-bibtex-value.h"
#include "xfbib-string.h"
#include "xfbib-integer.h"
#include "xfbib-window.h"
#include "xfbib-list-store.h"

static void foreach_bibtex_element(gpointer, gpointer);
static void foreach_field(gpointer, gpointer);
static void foreach_value(GList *, GString *);

/*
 * Opens a file and returns NULL if something went wrong otherwise it returns
 * the filename of the file that was parsed
 */
gchar *
xfbib_file_io_open(gchar *filename)
{
	XfbibState *state;
	XfbibBibtex *elements;
	GtkTreeModel *list_store_entries, *list_store_strings;
	gchar *contents, *name;
	gsize length;
	gint i, ret, n_entries, n_strings;
	GError *err = NULL;
	GtkWidget *dialog, *window;
	GtkFileFilter *filter;

	state = xfbib_state_new();

	window = xfbib_state_get_window(state);
	elements = xfbib_state_get_bibtex_elements(state);
	list_store_entries = gtk_tree_view_get_model(GTK_TREE_VIEW(xfbib_window_get_tree_view(XFBIB_WINDOW(window), TYPE_ENTRY)));
	list_store_strings = gtk_tree_view_get_model(GTK_TREE_VIEW(xfbib_window_get_tree_view(XFBIB_WINDOW(window), TYPE_STRING)));
	
	name = filename;

	if (name == NULL) {
		dialog = gtk_file_chooser_dialog_new(_("Open"),
				GTK_WINDOW(xfbib_state_get_window(state)),
				GTK_FILE_CHOOSER_ACTION_OPEN,
				GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
				GTK_STOCK_OPEN, GTK_RESPONSE_OK,
				NULL);

		filter = gtk_file_filter_new();
		gtk_file_filter_set_name(filter, _("BibTeX files"));
		gtk_file_filter_add_pattern(filter, "*.bib");
		gtk_file_chooser_add_filter(GTK_FILE_CHOOSER(dialog), filter);
		if ((ret = gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_OK)) {
			xfbib_list_store_clear(XFBIB_LIST_STORE(list_store_entries));
			xfbib_list_store_clear(XFBIB_LIST_STORE(list_store_strings));
			name = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(dialog));
		} else {
			gtk_widget_destroy(dialog);
			return NULL;
	 	}
		gtk_widget_destroy(dialog);
	}
	

	if (!g_file_get_contents(name, &contents, &length, &err)) {
		xfce_dialog_show_error(NULL, NULL, _("An error occurred when opening the file: %s\n"), err->message);
		g_error_free(err);
		return NULL;
	} else {
		n_entries = xfbib_bibtex_get_n(elements, TYPE_ENTRY);
		n_strings = xfbib_bibtex_get_n(elements, TYPE_STRING);
		/* Parse the file and update the tree view */
		if (xfbib_bibtex_parse(elements, contents)) {
			for (i = n_entries; i < xfbib_bibtex_get_n(elements, TYPE_ENTRY); i++) {
				xfbib_list_store_row_inserted(XFBIB_LIST_STORE(list_store_entries), i);
			}
			/* TODO: Update the string tree_view */
			for (i = n_strings; i < xfbib_bibtex_get_n(elements, TYPE_STRING); i++) {
				xfbib_list_store_row_inserted(XFBIB_LIST_STORE(list_store_strings), i);
			}
		}
	}
	
	return name;
}


gboolean
xfbib_file_io_save(gchar *name)
{
	FILE *fd;
	XfbibState *state;
	XfbibBibtex *elements;
	GString *str;

	if (name == NULL) {
		/* No filename, redirect to save as */
		return xfbib_file_io_save_as();
	}

	state = xfbib_state_new();
	elements = xfbib_state_get_bibtex_elements(state);

	str = g_string_new("");

	xfbib_bibtex_foreach(elements, foreach_bibtex_element, str);

	if ((fd = fopen(name, "w")) == NULL) {
		xfce_dialog_show_error(NULL, NULL, _("An error occurred when opening the file: %s\n"), strerror(errno));
		return FALSE;
	}

	if(fputs(str->str, fd) == EOF) {
		xfce_dialog_show_error(NULL, NULL, _("An error occurred when writing to the file: %s\n"), name);
		return FALSE;
	}

	g_string_free(str, TRUE);
	fclose(fd);
	return TRUE;
}


gboolean
xfbib_file_io_save_as(void)
{
	GtkWidget *dialog;
	gchar *name;
	XfbibState *state;
	g_print(N_("Save as\n"));
	

	state = xfbib_state_new();

	dialog = gtk_file_chooser_dialog_new(_("Save As"),
			GTK_WINDOW(xfbib_state_get_window(state)),
			GTK_FILE_CHOOSER_ACTION_SAVE,
			GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
			GTK_STOCK_SAVE, GTK_RESPONSE_OK,
			NULL);

	if (gtk_dialog_run(GTK_DIALOG(dialog)) == GTK_RESPONSE_OK) {
		name = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(dialog));
		if (xfbib_file_io_save(name)) {
			xfbib_state_set_filename(state, name);
			xfbib_window_set_title(XFBIB_WINDOW(xfbib_state_get_window(state)), name);	
		}
	}

	gtk_widget_destroy(dialog);
	return TRUE;
}


static void
foreach_bibtex_element(gpointer _obj, gpointer _str)
{
	GObject *obj = _obj;
	GString *str = _str;
	gchar separator;

	if (XFBIB_IS_BIBTEX_ENTRY(obj)) {
		separator = xfbib_bibtex_entry_get_separator(XFBIB_BIBTEX_ENTRY(obj));
		g_string_append_c(str, '@');
		g_string_append(str, xfbib_bibtex_entry_get_bibtype(XFBIB_BIBTEX_ENTRY(obj)));
		g_string_append_c(str, separator);
		g_string_append(str, xfbib_bibtex_entry_get_key(XFBIB_BIBTEX_ENTRY(obj)));
		g_string_append(str, ",\n");
		/* XfbibBibtexField */
		g_list_foreach(xfbib_bibtex_entry_get_fields(XFBIB_BIBTEX_ENTRY(obj)), foreach_field, str);
		g_string_append_c(str, (separator == '{') ? '}' : ')');
		g_string_append_c(str, '\n');
	} else if (XFBIB_IS_BIBTEX_STRING(obj)) {
		separator = xfbib_bibtex_string_get_separator(XFBIB_BIBTEX_STRING(obj));
		g_string_append(str, "@string ");
		g_string_append_c(str, separator);
		g_string_append(str, xfbib_bibtex_string_get_variable(XFBIB_BIBTEX_STRING(obj)));
		g_string_append(str, " = ");
		/* XfbibBibtexValue */
		foreach_value(xfbib_bibtex_value_get(xfbib_bibtex_string_get_value(XFBIB_BIBTEX_STRING(obj))), str);
		g_string_append_c(str, (separator == '{') ? '}' : ')');
		g_string_append_c(str, '\n');
	} else if (XFBIB_IS_BIBTEX_COMMENT(obj)) {
		g_string_append(str, xfbib_bibtex_comment_get(XFBIB_BIBTEX_COMMENT(obj)));
	} else if (XFBIB_IS_BIBTEX_PREAMBLE(obj)) {
		separator = xfbib_bibtex_preamble_get_separator(XFBIB_BIBTEX_PREAMBLE(obj));
		g_string_append(str, "@preamble ");
		g_string_append_c(str, separator);
		/* XfbibBibtexValue */
		foreach_value(xfbib_bibtex_value_get(xfbib_bibtex_preamble_get_value(XFBIB_BIBTEX_PREAMBLE(obj))), str);
		g_string_append_c(str, (separator == '{') ? '}' : ')');
		g_string_append_c(str, '\n');
	}
}


static void
foreach_field(gpointer obj, gpointer _str)
{
	XfbibBibtexField *field = obj;
	GString *str = _str;

	g_string_append_c(str, '\t');
	g_string_append(str, xfbib_bibtex_field_get_variable(XFBIB_BIBTEX_FIELD(obj)));
	g_string_append(str, " = ");
	/* XfbibBibtexValue */
	foreach_value(xfbib_bibtex_value_get(xfbib_bibtex_field_get_value(field)), str);
	g_string_append(str, ",\n");
}


static void
foreach_value(GList *list, GString *str)
{
	GList *obj;
	gchar separator;

	for (obj = g_list_first(list); obj != NULL; obj = g_list_next(obj)) {
		if (XFBIB_IS_STRING(obj->data)) {
			separator = xfbib_string_get_separator(XFBIB_STRING(obj->data));
			g_string_append_c(str, separator);
			g_string_append(str, xfbib_file_io_escape_commands(xfbib_string_get(XFBIB_STRING(obj->data))));
			g_string_append_c(str, (separator == '{') ? '}' : '"');
		} else if (XFBIB_IS_INTEGER(obj->data)) {
			g_string_append(str, xfbib_integer_get(XFBIB_INTEGER(obj->data)));
		} else if (XFBIB_IS_BIBTEX_STRING(obj->data)) {
			g_string_append(str, xfbib_bibtex_string_get_variable(XFBIB_BIBTEX_STRING(obj->data)));
		}
	
		if (g_list_last(list) != obj) {
			g_string_append(str, " # ");
		}
	}
}

gchar *
xfbib_file_io_escape_commands(gchar *str)
{
	int len, new_len, n, i;
	gchar *new_str;

	len = new_len = strlen(str) + 1;
	new_str = malloc(len);

	n = i = 0;
	while(n < len) {
		switch(str[n] & 0xff) {
			case '&':
			case '%':
				new_len++;
				new_str = realloc(new_str, new_len);
				new_str[i] = '\\';
				i++;
				new_str[i] = str[n];
				i++;
				n++;
				break;
			case 0xc3:
				n++;
				switch(str[n] & 0xff) {
					case 0x96:
						new_len += 3;
						new_str = realloc(new_str, new_len);
						new_str[i] = '\\';
						i++;
						new_str[i] = '\"';
						i++;
						new_str[i] = '{';
						i++;
						new_str[i] = 'O';
						i++;
						new_str[i] = '}';
						i++;
						n++;
						break;
					case 0xb6:
						new_len += 3;
						new_str = realloc(new_str, new_len);
						new_str[i] = '\\';
						i++;
						new_str[i] = '\"';
						i++;
						new_str[i] = '{';
						i++;
						new_str[i] = 'o';
						i++;
						new_str[i] = '}';
						i++;
						n++;
						break;
					case 0xa5:
						new_len += 2;
						new_str = realloc(new_str, new_len);
						new_str[i] = '\\';
						i++;
						new_str[i] = 'a';
						i++;
						new_str[i] = 'a';
						i++;
						new_str[i] = ' ';
						i++;
						n++;
						break;
					case 0x85:
						new_len += 2;
						new_str = realloc(new_str, new_len);
						new_str[i] = '\\';
						i++;
						new_str[i] = 'A';
						i++;
						new_str[i] = 'A';
						i++;
						new_str[i] = ' ';
						i++;
						n++;
						break;
					case 0xa4:
						new_len += 3;
						new_str = realloc(new_str, new_len);
						new_str[i] = '\\';
						i++;
						new_str[i] = '\"';
						i++;
						new_str[i] = '{';
						i++;
						new_str[i] = 'a';
						i++;
						new_str[i] = '}';
						i++;
						n++;
						break;
					case 0x84:
						new_len += 3;
						new_str = realloc(new_str, new_len);
						new_str[i] = '\\';
						i++;
						new_str[i] = '\"';
						i++;
						new_str[i] = '{';
						i++;
						new_str[i] = 'A';
						i++;
						new_str[i] = '}';
						i++;
						n++;
						break;
				}
				break;
			default:
				new_str[i] = str[n];
				i++;
				n++;
		}
	}
	return new_str;
}
